/*
  Implementation of AVL trees
*/

compile_mode :pop11 +strict;

lconstant AVL_DEBUG = true;

define search(tree, key) -> el;
lvars key, tree, el = tree;
  while el /= [] do
    if el(1) = key then return endif;
    if el(1) > key then el(2) else el(3) endif -> el;
  endwhile;
  return;
enddefine;

define height(tree) -> result;
  lvars tree, result;
  if tree = [] then 
    0
  else
    max(height(tree(2)), height(tree(3))) + 1
  endif -> result;
enddefine;

define balance1_?(tree) -> result;
  lvars tree, result, hl, hp, result = true;
  if tree = [] then return; endif;
  height(tree(2)) -> hl;
  height(tree(3)) -> hp;
  if abs(hl - hp) > 1 then 
     false -> result;
     tree ==>
  else
     if hl - hp /= tree(4) then
        tree ==>
        mishap(0, 'Bad balance indicator');
     endif;
  endif;
enddefine;

define balance_?(tree) -> result;
  lvars tree, result = true;
  if tree = [] then return; endif;
  if balance_?(tree(2)) and balance_?(tree(3)) and balance1_?(tree) then
    return;
  else
   false -> result;
  endif;
enddefine;

define check_balance(tree);
    lvars tree;
    if not(balance_?(tree)) then
      mishap(0, 'Inbalance');
    endif;
enddefine;

define copyavltree(tree) -> result;
  lvars tree, result = [];
  if tree /= [] then
    [^(tree(1)) ^(copyavltree(tree(2))) ^(copyavltree(tree(3))) ^(tree(4))] -> result;
  endif;
enddefine;


/*
   procedura balance2.
   Goal: re-balance a tree
   Inputs:
     tree   inbalanced tree which both subtrees are in balance.
     i      index of inbalanced tree

   Cases:
    Single rotation
         tree -             -->        a 0
      a-   b                       c     tree 0
    c   d                               d   b

   l+2 l+1 l                      l+1  l+1  l+1 
    Double rotation
          tree +           -->        d 0
      a-         b               a s'      tree s''
   c      d s                  c   e     f   b
        e   f

   l  l+2 l+1    l             l  l+1    l   l+1

   Remark: balaced tree as an argument is an error 
*/

define balance2(tree, i) -> result;
    lvars tree, i, a = tree(i), d = a(5 - i), s, result;
#_IF AVL_DEBUG
    lvars otree = copyavltree(tree);
#_ENDIF
    if i = 2 then 1 else -1 endif -> s;
    if s /= -a(4) then
      ;;; Single rotation
      tree -> a(5 - i);
      a -> result;
      d -> tree(i);
      a(4) - s -> a(4);
      - a(4) -> tree(4);
    else
      ;;; Double rotation
      d(i) -> a(5 - i);
      d(5 - i) -> tree(i);
      a -> d(i);
      tree -> d(5 - i);
      if d(4) = s then (0, -s) else (-d(4), 0) endif -> (a(4), tree(4));
      d -> result;
      0 -> result(4);
    endif;
#_IF AVL_DEBUG
    if not(balance_?(result)) then
      printf(i, 'i = %p, tree:\n');
      otree ==>
      printf('result:\n');
      result ==>
      mishap(0, 'Inbalance');
    endif;
#_ENDIF
enddefine;


;;; [5 [2 [1 [] [] 0] [] 1] [] 1]
;;; [5 [2 [1 [] [] 0] [4 [3 [] [] 0] [] 1] -1] [6 [] [] 0] 1]


/* 
   Procedure balance1.
   Goal: update the balance indicator and if needed re-balance the
        tree (calling balance2).
   Inputs:
   tree   tree such that one of its direct subtree has changed its height
   i    index, showing in which direction of the change, that is
        index of the subtree which increased its height or complement
        of the index of the subtree which decreased its height
   s    1 for i = 2 , -1 for i = 3
*/

define balance1(tree, i, s) -> (result, delta);
    lvars tree, i, s, delta = 1, result = tree;
#_IF AVL_DEBUG
    lvars height_orig =
      max(height(tree(i)) + 1, height(tree(5 - i)) + 2);
    lvars otree = copyavltree(tree);

    if i = 2 and s /= 1 or i = 3 and s /= -1 then
      printf('***********************\n');
    endif;
    lvars lh = height(tree(2)), rh = height(tree(3));
    lvars adelta = lh - rh;
    if abs(lh - rh) /= s*adelta then
      printf(i, 'i = %p, ');
      printf(s, 's = %p\n');
      tree ==>
      mishap(0, 'Bad s');
    endif;
    if abs(lh - rh) > 2 then
      printf('tree:\n');
      tree ==>
      mishap(0, 'Bad tree');
   endif;
   if lh - rh /= tree(4) + s then
     printf(i, 'i = %p, ');
     printf(s, 's = %p\n');
     printf('tree:\n');
     tree ==>
     mishap(0, 'Bad info about balance');
   endif;
#_ENDIF
    if tree(4) = 0 then
      s -> tree(4);
    elseif s = -tree(4) then
      0 -> tree(4);
      0 -> delta;
    else
      balance2(tree, i) -> result;
      abs(result(4)) -> delta;
    endif;
#_IF AVL_DEBUG
    if height_orig - height(result) /= (1 - delta) then
      printf('tree:\n');
      otree ==>
      mishap(0, 'Bad delta');
    endif;
    check_balance(result);
#_ENDIF
enddefine;


define addtree1(tree, key) -> (result, delta);
    lvars tree, key, result = tree, delta = 0;
    if tree == [] then
        1 -> delta;
        [^key [] [] 0] -> result;
        return;
    endif;
    
    if tree(1) = key then
            return;
    endif;
   
    lvars i, s;
    if tree(1) > key then (2, 1) else (3, -1) endif -> (i, s);
    addtree1(tree(i), key) -> (tree(i), delta);

    if delta = 0 then return; endif;
    balance1(tree, i, s) -> (result, delta);
enddefine;


define addtree(tree, key) -> result;
    lvars tree, key, result;
    addtree1(tree, key) ->; -> result;
enddefine;

;;; [5 [2 [1 [] [] 0] [] 1] [] 1]
;;; [5 [2 [1 [] [] 0] [4 [3 [] [] 0] [] 1] -1] [6 [] [] 0] 1]

/* 
   Procedure balance1r.
   Goal: call balance1 and then correct delta for the purpose of
         removal.
*/


define balance1r(tree, i, s) -> (result, delta);
lvars tree, i, s, result, delta;
  balance1(tree, i, s) -> (result, delta);
  1 - delta -> delta;
enddefine;

/*
  Procedure remtree2.
  Goal: removes the rightmost node of the tree and returns it as r_el
        transformed tree is returned as r_t 
*/
define remtree2(tree) -> (r_el, r_t, d1);
  lvars tree, r_el, r_t = tree;
#_IF AVL_DEBUG
  lvars height_orig = height(tree);
  lvars otree = copyavltree(tree);
#_ENDIF
  if tree(3) = [] then
    tree(2) -> r_t;
    tree -> r_el;
    1 -> d1;
  else
    remtree2(tree(3)) -> (r_el, tree(3), d1);
    if d1 /= 0 then 
      balance1r(tree, 2, 1) -> (r_t, d1);
    endif;
  endif;
#_IF AVL_DEBUG
  if height_orig - height(r_t) /= d1 then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Bad d1');
  endif;
#_ENDIF
enddefine;

/*
  Procedure remtree0.
  Goal: remove the root of the tree
*/
define remtree0(tree) -> (result, delta);
  lvars tree, result, delta, tmp;
#_IF AVL_DEBUG
  lvars height_orig = height(tree);
  lvars otree = copyavltree(tree);
#_ENDIF
  remtree2(tree(2)) -> (result, tmp, delta);
  tmp -> result(2);
  tree(3) -> result(3);
  tree(4) -> result(4);
  if delta /= 0 then
      balance1r(result, 3, -1) -> (result, delta);
  endif;
#_IF AVL_DEBUG
  if height_orig - height(result) /= delta then
      printf('tree:\n');
      otree ==>
      mishap(0, 'Bad delta');
  endif;
#_ENDIF      
enddefine;

vars remtree1;

define remtree_rec(tree, key) -> (result, delta);
  lvars tree, key, result = tree, delta;
  lvars i, s;
#_IF AVL_DEBUG
  lvars height_orig = height(tree);
  lvars otree = copyavltree(tree);
#_ENDIF
  if tree(1) > key then (2, -1) else (3, 1) endif -> (i, s);
#_IF AVL_DEBUG
  lvars ohi = height(tree(i));
#_ENDIF
  remtree1(tree(i), key) -> (tree(i), delta);
#_IF AVL_DEBUG
  if ohi - height(tree(i)) /= delta then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Bad delta (after remtree1)');
  endif;
#_ENDIF
  if delta /= 0 then
    balance1r(tree, 5 - i, s) -> (result, delta);
  endif;
#_IF AVL_DEBUG
  if height_orig - height(result) /= delta then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Bad delta');
  endif;
  if not(balance_?(result)) then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Inbalance');
  endif;
#_ENDIF
enddefine;

/*
  Procedura remtree1.
  Goal: removes key from the tree, handling directly trivial
        cases, otherwise calls remtree_rec (the recursive case)
        or remtree0 (for inner nodes).
       At the end result contains the resulting tree and
       delta contains change of the height (0 - no change,
       1 - height decreased by 1)
*/

define remtree1(tree, key) -> (result, delta);
  lvars tree, key, result = tree, delta = 0;
#_IF AVL_DEBUG
  lvars height_orig = height(tree);
  lvars otree = copyavltree(tree);
#_ENDIF
  if tree = [] then return endif;
  if tree(1) = key then
    if tree(2) = [] then
      tree(3) -> result;
      1 -> delta;
    elseif tree(3) = [] then
      tree(2) -> result;
      1 -> delta;
    else
      remtree0(tree) -> (result, delta);
    endif;
  else
    remtree_rec(tree, key) -> (result, delta);
  endif;
#_IF AVL_DEBUG
  if height_orig - height(result) /= delta then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Bad delta');
  endif;
  if not(balance_?(result)) then
    printf('tree:\n');
    otree ==>
    mishap(0, 'Inbalance');
  endif;
#_ENDIF
enddefine;

define remtree(dz, key) -> result;
  lvars dz, key, result;
  remtree1(dz, key) -> ; -> result;
  check_balance(result);
enddefine;
 
vars a, i;

[] -> a;
/* Tests on random trees */

'random(200)' ==>
lvars j;
for j from 1 to 100 do
  for i from 1 to 100 do addtree(a, random(200)) -> a; endfor;
  for i from 1 to 100 do remtree(a, random(200)) -> a; endfor;
  printf('.\n');
endfor;

/*
;;; a ==>
balance_?(a) ==>
for i from 1 to 20000 do remtree(a, random(40000)) -> a; endfor;

for i from 1 to 40000 do remtree(a, i) -> a; endfor;

a ==>
*/

/*
[] -> a;
for i from 1 to 100 do addtree(a, i) -> a; endfor;
a ==>
balance_?(a) ==>


[1211 [1152 [1147 [] [] 0]
               [1199 [1153 [] [] 0] [1208 [] [] 0] 0]
               -1]
         [1252 [1238 [1218 [1214 [] [] 0] [] 1]
                     [1240 [1239 [] [] 0] [1251 [] [] 0] 0]
                     0]
               [1290 [1280 [1279 [] [] 0] [] 1]
                     [1309 [1306 [] [] 0] [] 1]
                     0]
               0]
         -1] -> a;
*/

balance_?(a) => ;
remtree(a, 1238) -> a;

for i from 1 to 100 do remtree(a, i) -> a; /* a ==> */ endfor;
a ==>
balance_?(a) ==>


/*
[16 [8 [4 [2 [1 [] [] 0] [3 [] [] 0] 0]
                     [6 [5 [] [] 0] [7 [] [] 0] 0]
                     0]
                  [12 [10 [9 [] [] 0] [11 [] [] 0] 0]
                      [14 [13 [] [] 0] [15 [] [] 0] 0]
                      0]
                  0]
               [24 [20 [18 [17 [] [] 0] [19 [] [] 0] 0]
                       [22 [21 [] [] 0] [23 [] [] 0] 0]
                       0]
                   [28 [26 [25 [] [] 0] [27 [] [] 0] 0]
                       [30 [29 [] [] 0] [31 [] [] 0] 0]
                       0]
                   0]
               0] -> a;

for i from 1 to 11 do remtree(a, i) -> a;  /* a ==> */ endfor;
a ==>
*/

[16 [14 [12 [] [13 [] [] 0] -1] [15 [] [] 0] 1]
       [24 [20 [18 [17 [] [] 0] [19 [] [] 0] 0]
               [22 [21 [] [] 0] [23 [] [] 0] 0]
               0]
           [28 [26 [25 [] [] 0] [27 [] [] 0] 0]
               [30 [29 [] [] 0] [31 [] [] 0] 0]
               0]
           0]
       -1] -> a;
a ==>
check_balance(a);
remtree(a, 12);
/* remtree(a, 14); 
*/

/*
[16 [14 [] [15 [] [] 0] -1]
       [24 [20 [18 [17 [] [] 0] [19 [] [] 0] 0]
               [22 [21 [] [] 0] [23 [] [] 0] 0]
               0]
           [28 [26 [25 [] [] 0] [27 [] [] 0] 0]
               [30 [29 [] [] 0] [31 [] [] 0] 0]
               0]
           0]
       -2] -> a;
a ==>
remtree(a, 14);
*/


[] -> a;
for i from 2000 by -1 to 1 do addtree(a, i) -> a; endfor;
balance_?(a) ==>

[] -> a;
for i from 1 by 3 to 2000 do remtree(a, i) -> a; endfor;
for i from 1 by 3 to 2000 do remtree(a, i) -> a; endfor;
for i from 2 by 3 to 2000 do remtree(a, i) -> a; endfor;
for i from 3 by 3 to 2000 do remtree(a, i) -> a; endfor;
balance_?(a) ==>

