compile_mode :pop11 +strict;

/*
   Parser for a small subset of Pop11. Slightly expanded compared to
   parse1.p: we allow all four arithmetic operators. Also, we allow
   a complete set of comparison operators. To simplify the resulting
   tree we treat each operator a binary operator.
*/

vars itemrep;
incharitem(charin) -> itemrep;

;;; Biecy symbol
vars sym; 

define get_sym();
    itemrep() -> sym;
enddefine;

define expect(x);
    lvars x;
    if x /= sym then
        printf(x, 'Error, expected %p\n');
        mishap(sym, 1, 'Syntax error');
    endif;
    get_sym();
enddefine;


lconstant sentence_stopers = [) endif endwhile else + * - / 
                              = /= < > >= <= ^(termin)];
lconstant res_list = [if then while do ( ^^(sentence_stopers)];

lconstant sentence_stop = newproperty(
  maplist(sentence_stopers, procedure(x); [^x ^(true)]; endprocedure),
  20, false, "perm");

lconstant reserved = newproperty(
  maplist(res_list, procedure(x); [^x ^(true)]; endprocedure),
  20, false, "perm");


/*
  The parser proper. Each nonterminal (rule) has a corresponding
  procedure responsible to recognize it.

  Remark: all words, exect for reserved words (which we remember
  in the hash table reserved) are treated as variable names.

*/

/*
sentences: sentence
   | sentences sentence
   ;
*/

define do_sentences() -> result;
    lvars result = [];
    while true do
        conspair(do_sentence(), result) -> result;
        quitif(sentence_stop(sym));
    endwhile;
    if back(result) /= [] then
        [sentences ^(rev(result))] -> result;
    else
        front(result) -> result;
    endif;
enddefine;

/*
sentence: expr ";"
   | expr "->" word ";"
   ;
*/

define do_sentence() -> result;
    do_expr() -> result;
    if sym = "->" then
        get_sym();
        [-> ^(sym) ^(result)] -> result;
        get_sym();();
    endif;
    expect(";");
enddefine;

/*
expr: "if" expr "then" sentences "else" sentences "endif"
   | "while" expr "do" sentences "endwhile"
   | ar_expr "=" ar_expr
....
   | ar_expr ">=" ar_expr
   | ar_expr
   ;
*/

define do_expr() -> result;
    lvars war, zd1, zd2, expr2;
    if sym = "if" then
        get_sym();
        do_expr() -> war;
        expect("then");
        do_sentences() -> zd1;
        expect("else");
        do_sentences() -> zd2;
        expect("endif");
        [if ^war ^zd1 ^zd2] -> result; 
    elseif sym = "while" then
        get_sym();
        do_expr() -> war;
        expect("do");
        do_sentences() -> zd1;
        expect("endwhile");
        [while ^war ^zd1] -> result;
    else
        do_ar_expr() -> result;
        if sym = "=" or sym = "/=" or sym = "<" or sym = "<=" or
           sym = ">" or sym = ">=" then
            lvars operator = sym;
            get_sym();
            do_ar_expr() -> expr2;
            [operator ^result ^expr2] -> result;
        endif;
    endif;
enddefine;

/*
ar_expr: term
   | ar_expr "+" term
   | ar_expr "-" term
   ;
*/

define do_ar_expr() -> result;
    lvars result = [], operator, term;
    while true do
        if result = [] then
            do_term() -> result;
        else
            [^operator ^result ^(do_term())] -> result;
        endif;
        quitif(sym /= "+" and sym /= "-");
        sym -> operator;
        get_sym();
    endwhile;
enddefine;

/*
term: factor
   | term "*" factor
   | term "/" factor
   ;
*/

define do_term() -> result;
    lvars result = [], operator;
    while true do
        if result = [] then
            do_factor() -> result;
        else
            [^operator ^result ^(do_factor())] -> result;
        endif;
        quitif(sym /= "*" and sym /= "/");
        sym -> operator;
        get_sym();
    endwhile;
enddefine;

/*
factor: word
   | constant
   | "(" sentences ")"
   ;
*/

define do_factor() -> result;
    if sym = "(" then
        get_sym();
        do_sentences() -> result;
        expect(")");
    elseif isnumber(sym) then
        sym -> result;
        get_sym();
    else
        if reserved(sym) then
            mishap(sym, 1, 'Bd skadni, nieoczekiwany symbol');
        endif;
        sym -> result;
        get_sym();
    endif;
enddefine;

get_sym();
do_sentences() ==>
